--[[
 * sota_event_system.lua
 * Copyright (C) 2022 Garritt Grandberg (aka Tirrag/peace-n-love)

   sota_event_system is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   sota_event_system is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   
	----------------------------------------------------------------------------
	changes
	----------------------------------------------------------------------------
	
	v1.02 - added ShroudOnSoundLoaded as an event that will notify when a sound has been loaded
	v1.03 - added ShroudOnStopSound as an event that will notify when a sound has been been stopped
	v1.04 - added ShroudOnSceneChange and ShroudOnUIActiveChange events
	v1.05 - added missing isStart parameters to the ShroudOnPlayFX and ShroudOnPlaySound events
]]

local ScriptName = "SotA Event System (SES) Addon";
local Version = "1.05";
local CreatorName = "Garritt Grandberg (aka Tirrag/peace-n-love)";
local Description = "Event handler system to allow notification of the ShroudOn events within addons and modules that do not have access to the ShroudOn event functions directly.  Also supports custom events and allows notification across addons enabling cross-addon communication.  Loosely based on event handler system from libsota library.";

SES = {
	utility = {
		-- https://stackoverflow.com/questions/640642/how-do-you-copy-a-lua-table-by-value
		deepcopy = function(orig)
			local orig_type = type(orig)
			local copy
			if orig_type == 'table' then
				copy = {}
				for orig_key, orig_value in next, orig, nil do
					copy[SES.utility.deepcopy(orig_key)] = SES.utility.deepcopy(orig_value)
				end
				setmetatable(copy, SES.utility.deepcopy(getmetatable(orig)))
			else -- number, string, boolean, etc
				copy = orig
			end
			return copy
		end,

		-- based off of libsota ui.module.add
		loadmodule = function(modulename)
			local part = {}
			
			for p in modulename:gmatch("[^\\/.]+") do part[#part + 1] = p end
			
			if part[#part] == "lua" then part[#part] = nil end
			
			modulename = table.concat(part, ".")
			
			local filename = table.concat(part, "/")..".lua"
			local f = nil
			for _,inc in next, {"/","/lib/","/share/","/bin/","/include/"} do
				f = io.open(ShroudLuaPath..inc..filename)
				if f then break end
			end
			if not f then
				print("loadmodule: '",filename,"' not found")
				return nil
			end
			local d = f:read("*a")
			f:close()
			local m,e = loadsafe(d, modulename, "bt", _G)
			if not m then
				print("loadmodule: '",modulename,"': "..e)
				return nil
			end
			
			return m
		end,
	},
	_is_init = false, 
	_is_enabled = false,
	_handlers = {
		list = {},
		-- adds a callback for a handler
		-- returns index of new handler
		add = function(eventName, callback)
			-- if this is the first callback, setup the handler callback list
			if (SES._handlers.list[eventName] == nil) then
				SES._handlers.list[eventName] = {
					_is_muted = false,
					callbacks = {}
				}
			end
		
			local callbackIndex = #SES._handlers.list[eventName].callbacks + 1

			SES._handlers.list[eventName].callbacks[callbackIndex] = {
				_index = callbackIndex,
				eventName = eventName,
				callback = callback,
			}
			
			-- if this is the special ShroudOnStart event and the event system has already initialized (it has already fired) then fire this now
			-- this can happen when using the lua addon manager interface to turn on a lua addon
			if (eventName == "ShroudOnStart" and SES._is_init == true) then
				callback()
			end
						
			return callbackIndex	
		end,
		-- removes a handler for an event by callbackIndex or all handlers if no callbackIndex is passed
		remove = function(eventName, callbackIndex)
			if (SES._handlers.list[eventName] != nil) then
				-- if callbackIndex is nil then clear all callbacks
				if (callbackIndex == nil) then
					for callbackIndex, callbackData in pairs(SES._handlers.list[eventName].callbacks) do
					
						if (callbackData != nil and callbackData.callback != nil) then 
							callbackData.callback = nil							
						end
					
						SES._handlers.list[eventName].callbacks[callbackIndex] = nil
					end
					SES._handlers.list[eventName] = nil
				else
					SES._handlers.list[eventName].callbacks[callbackIndex] = nil
				end
			end
		end,
		-- invokes all callbacks for an event
		invoke = function(eventName, ...)
			-- if the system has not initialized or there are no callbacks for the event, skip
			if not SES._is_init or not SES._is_enabled or SES._handlers.list[eventName] == nil then return end
			if SES._handlers.list[eventName]._is_muted == true then return end

			for callbackIndex, callbackData in pairs(SES._handlers.list[eventName].callbacks) do
				if (callbackData != nil and callbackData.callback != nil) then 
					callbackData.callback(...)
				end
			end
		end,
		mute = function(eventName, isMuted)
			if (SES._handlers.list[eventName] != nil) then
				SES._handlers.list[eventName]._is_muted = isMuted
			end
		end,
	},
	on = function (eventName, callback) return SES._handlers.add(eventName, callback) end,
	off = function (eventName, callbackIndex) return SES._handlers.remove(eventName, callbackIndex) end,
	invoke = function (eventName, ...) SES._handlers.invoke(eventName, ...) end,
	mute = function (eventName) SES._handlers.mute(eventName, true) end,
	unmute = function (eventName) SES._handlers.mute(eventName, false) end,
	enabled = function (isEnabled) 
		SES._is_enabled = isEnabled 
		if (SES._is_enabled) then 
			ShroudConsoleLog(ScriptName .. " v" .. Version .. ": [00FF00]Enabled[-]")
		else 
			ShroudConsoleLog(ScriptName .. " v" .. Version .. ": [FF0000]Disabled[-]")
		end
	end,
}

function ShroudOnStart()
end

local lastSoundPoll = nil
local lastSoundList = {}
local lastChannelList = {}
local soundPollInterval = 0.5

local lastScenePoll = nil
local scenePollInterval = 0.5
local lastRawSceneName = nil

local lastUIActivePoll = nil
local uiActivePollInterval = 0.5
local lastUIActive = nil

function ShroudOnUpdate()
	if not SES._is_init then
		if not ShroudServerTime then return end -- shroud Api not ready, yet

		SES._is_init = true

		ShroudConsoleLog(ScriptName .. " v" .. Version .. ": [00FF00]Ready[-]")

		SES.enabled(true)

		SES.invoke("ShroudOnStart")
	else
		SES.invoke("ShroudOnUpdate")
		
		local ts = os.time()
		
		if (lastScenePoll == nil or (ts - lastScenePoll) >= scenePollInterval) then 
			local currentRawSceneName = ShroudGetCurrentSceneNameRaw()
		
			if (lastRawSceneName != currentRawSceneName) then
				SES.invoke("ShroudOnSceneChange", ShroudGetCurrentSceneName(), currentRawSceneName)
				lastRawSceneName = currentRawSceneName
			end
			
			lastScenePoll = ts
		end

		if (lastUIActivePoll == nil or (ts - lastUIActivePoll) >= uiActivePollInterval) then 
			local currentUIActive = ShroudIsUIActive()
		
			if (lastUIActive != currentUIActive) then
				SES.invoke("ShroudOnUIActiveChange", currentUIActive)
				lastUIActive = currentUIActive
			end
			
			lastUIActivePoll = ts
		end
		
		if (lastSoundPoll != nil and (ts - lastSoundPoll) < soundPollInterval) then return end

		-- check for sounds being loaded
		local currentSoundList = ShroudListSound()	

		for soundID, soundName in pairs(currentSoundList) do
			local found = false
			for lastSoundID, lastSoundName in pairs(lastSoundList) do
				if (soundName == lastSoundName) then
					found = true
					break
				end
			end
			if (found == false) then 
				SES.invoke("ShroudOnSoundLoaded", soundID, soundName)
			end
		end
		
		lastSoundList = currentSoundList
		
		local currentChannelList = {}
		
		-- check for sounds stopped playing (5 channels)
		
		-- get the list of current sounds
		for soundChannel = 1, 5, 1 do
			local playingSoundName = ShroudIsChannelPlaying(soundChannel)

			if (playingSoundName != "" and playingSoundName != -1) then
				currentChannelList[playingSoundName] = soundChannel
			end			
		end
		
		-- compare against last list
		for lastSoundName, lastSoundChannel in pairs(lastChannelList) do
			local isStillPlaying = false
			for currentSoundName, currentSoundChannel in pairs(currentChannelList) do 
				if (lastSoundName == currentSoundName and lastSoundChannel == currentSoundChannel) then
					isStillPlaying = true
					break
				end
			end
			
			-- if no longer playing, notify
			if (isStillPlaying == false) then
				SES.invoke("ShroudOnStopSound", lastSoundName, lastSoundChannel)
			end			
		end

		lastChannelList = currentChannelList

		lastSoundPoll = ts
	end
end

function ShroudOnGUI()
	SES.invoke("ShroudOnGUI")
end

function ShroudOnConsoleInput(channel, sender, message)
	SES.invoke("ShroudOnConsoleInput", channel, sender, message)
end

function ShroudOnMouseClick(objectID, objectKind)
	SES.invoke("ShroudOnMouseClick" .. tostring(objectID) .. tostring(objectKind))
	SES.invoke("ShroudOnMouseClick", objectID, objectKind)
end

function ShroudOnMouseOver(objectID, objectKind)
	SES.invoke("ShroudOnMouseOver" .. tostring(objectID) .. tostring(objectKind))
	SES.invoke("ShroudOnMouseOver", objectID, objectKind)
end

function ShroudOnMouseOut(objectID, objectKind)
	SES.invoke("ShroudOnMouseOut" .. tostring(objectID) .. tostring(objectKind))
	SES.invoke("ShroudOnMouseOut", objectID, objectKind)
end

function ShroudOnSceneLoaded(sceneName)
	SES.invoke("ShroudOnSceneLoaded", sceneName)
end

function ShroudOnSceneUnloaded()
	SES.invoke("ShroudOnSceneUnloaded")
end

function ShroudOnLogout()
	SES.invoke("ShroudOnLogout")
end

function ShroudOnInputChange(objectID, text)
	SES.invoke("ShroudOnInputChange" .. tostring(objectID), text)
	SES.invoke("ShroudOnInputChange", objectID, text)
end

function ShroudOnToggleChange(objectID, isOn)
	SES.invoke("ShroudOnToggleChange" .. tostring(objectID), isOn)
	SES.invoke("ShroudOnToggleChange", objectID, isOn)
end

function ShroudOnPlayFX(fxName, isStart)
	SES.invoke("ShroudOnPlayFX", fxName, isStart)
end

function ShroudOnPlaySound(soundName, isStart)
	SES.invoke("ShroudOnPlaySound", soundName, isStart)
end

function ShroudOnExperienceGain(type, amount)
	SES.invoke("ShroudOnExperienceGain", type, amount)
end

function ShroudOnDisableScript()
	SES.invoke("ShroudOnDisableScript")
	SES.enabled(false)
	ShroudConsoleLog(ScriptName .. ": [FF0000]Script Disabled[-]")
end

